/*
 * Alessandro Rubini for CERN, 2011 -- public domain
 */

/*
 * This is the startup thing for hosted environments. It
 * defines main and then calls the main loop.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>


#include <ppsi/ppsi.h>
#include "posix.h"

CONST_VERBOSITY int pp_diag_verbosity = 0;

/* FIXME: make MAX_LINKS and conf_path definable at compile time */
#define MAX_LINKS 32
#define CONF_PATH "/etc/ppsi.conf"

int main(int argc, char **argv)
{
	struct pp_globals *ppg;
	struct pp_instance *ppi;
	char *ifname;
	char tmp[16];
	int i = 0;
	struct stat conf_fs;
	char *conf_buf;
	int conf_fd;
	int conf_len = 0;

	setbuf(stdout, NULL);

	/* We are hosted, so we can allocate */
	ppg = calloc(1, sizeof(*ppg));
	if (!ppg)
		exit(__LINE__);

	ppg->max_links = MAX_LINKS;
	ppg->links = calloc(ppg->max_links, sizeof(struct pp_link));

	if ((stat(CONF_PATH, &conf_fs) < 0) ||
	    (conf_fd = open(CONF_PATH, O_RDONLY) < 0)) {
		pp_printf("Warning: could not open %s, default to one-link built-in "
					"config\n", CONF_PATH);
		conf_buf = "link 0\niface eth0";
		conf_len = strlen(conf_buf);
	}
	else {
		int r = 0, next_r;
		conf_buf = calloc(1, conf_fs.st_size) + 1;

		do {
			next_r = strlen(conf_buf) - r;
			r = read(conf_fd, &conf_buf[conf_len], next_r);
			if (r <= 0)
				break;
		} while (conf_len < conf_fs.st_size);

		close(conf_fd);
	}

	pp_parse_conf(ppg, conf_buf, conf_len);

	ppg->pp_instances = calloc(ppg->nlinks, sizeof(struct pp_instance));

	if (!ppg->pp_instances)
		exit(__LINE__);

	for (; i < ppg->nlinks; i++) {

		ifname = ppg->links[i].iface_name;

		if (!ifname) {
			sprintf(tmp, "eth%d", i);
			ifname = tmp;
		}

		ppi = &ppg->pp_instances[i];
		ppi->glbs = ppg;
		/* FIXME set ppi proto and ext as defined in its pp_link */

		/* FIXME check all of these calloc's, since some stuff will be
		 * part of pp_globals */
		GLBS(ppi)->defaultDS = calloc(1, sizeof(*GLBS(ppi)->defaultDS));
		GLBS(ppi)->currentDS = calloc(1, sizeof(*GLBS(ppi)->currentDS));
		GLBS(ppi)->parentDS = calloc(1, sizeof(*GLBS(ppi)->parentDS));
		ppi->portDS = calloc(1, sizeof(*ppi->portDS));
		GLBS(ppi)->timePropertiesDS = calloc(1, sizeof(*GLBS(ppi)->timePropertiesDS));
		GLBS(ppi)->servo = calloc(1, sizeof(*GLBS(ppi)->servo));
		ppi->arch_data = calloc(1, sizeof(struct posix_arch_data));

		ppi->n_ops = &posix_net_ops;
		ppi->t_ops = &posix_time_ops;

		if ((!GLBS(ppi)->defaultDS) || (!GLBS(ppi)->currentDS) || (!GLBS(ppi)->parentDS)
			|| (!ppi->portDS) || (!GLBS(ppi)->timePropertiesDS)
			|| (!GLBS(ppi)->frgn_master) || (!ppi->arch_data)
		)
			exit(__LINE__);
	}

	/* FIXME temporary workaround to make the first interface work as in the past */
	if (ppg->nlinks == 1) {
		struct pp_instance *ppi = &ppg->pp_instances[0];
		ppi->iface_name = strdup(ifname);
		ppi->ethernet_mode = PP_DEFAULT_ETHERNET_MODE;
		pp_open_instance(ppg, NULL);
		if (pp_parse_cmdline(ppi, argc, argv) != 0)
			return -1;
	}

	posix_main_loop(ppg);
	return 0; /* never reached */
}
